package de.kif.backend

import com.uchuhimo.konf.Config
import com.uchuhimo.konf.ConfigSpec
import com.uchuhimo.konf.Item
import java.io.FileNotFoundException
import java.nio.file.Files
import java.nio.file.Paths
import java.text.SimpleDateFormat
import java.time.ZoneId
import java.util.*
import kotlin.reflect.KProperty

object Configuration {

    private val config: Config

    private class ConfigDelegate<T>(val item: Item<T>) {
        operator fun getValue(thisRef: Any?, property: KProperty<*>): T {
            return config[item]
        }
    }

    private fun <T> c(item: Item<T>) = ConfigDelegate(item)

    private object ServerSpec : ConfigSpec("server") {
        val host by required<String>()
        val port by required<Int>()
        val debug by required<Boolean>()
        val cPrefix by required<String>("prefix")
    }

    object Server {
        val host by c(ServerSpec.host)
        val port by c(ServerSpec.port)
        val debug by c(ServerSpec.debug)
        val prefix by c(ServerSpec.cPrefix)
    }

    private object PathSpec : ConfigSpec("path") {
        val web by required<String>()
        val sessions by required<String>()
        val uploads by required<String>()
        val database by required<String>()
        val announcement by required<String>()
        val backup by required<String>()
    }

    object Path {
        val web by c(PathSpec.web)
        val webPath: java.nio.file.Path by lazy { Paths.get(web).toAbsolutePath() }

        val sessions by c(PathSpec.sessions)
        val sessionsPath: java.nio.file.Path by lazy { Paths.get(sessions).toAbsolutePath() }

        val uploads by c(PathSpec.uploads)
        val uploadsPath: java.nio.file.Path by lazy { Paths.get(uploads).toAbsolutePath() }

        val database by c(PathSpec.database)
        val databasePath: java.nio.file.Path by lazy { Paths.get(database).toAbsolutePath() }

        val announcement by c(PathSpec.announcement)
        val announcementPath: java.nio.file.Path by lazy { Paths.get(announcement).toAbsolutePath() }

        val backup by c(PathSpec.backup)
        val backupPath: java.nio.file.Path by lazy { Paths.get(backup).toAbsolutePath() }
    }

    private object ScheduleSpec : ConfigSpec("schedule") {
        val reference by required<String>()
        val offset by required<Long>()
        val wallStart by required<Int>("wall_start")
    }

    object Schedule {
        val reference by c(ScheduleSpec.reference)
        val referenceDate: Date by lazy {
            val sdf = SimpleDateFormat("yyyy-MM-dd")
            sdf.timeZone = TimeZone.getTimeZone(ZoneId.of("UTC"))
            val calendar = Calendar.getInstance()
            calendar.time = sdf.parse(reference)
            calendar.set(Calendar.HOUR_OF_DAY, 0)
            calendar.set(Calendar.MINUTE, 0)
            calendar.set(Calendar.SECOND, 0)
            calendar.set(Calendar.MILLISECOND, 0)
            calendar.time
        }

        val offset by c(ScheduleSpec.offset)
        val wallStart by c(ScheduleSpec.wallStart)
    }

    private object SecuritySpec : ConfigSpec("security") {
        val sessionName by required<String>("session_name")
        val signKey by required<String>("sign_key")
    }

    object Security {
        val sessionName by c(SecuritySpec.sessionName)
        val signKey by c(SecuritySpec.signKey)
    }

    private object GeneralSpec : ConfigSpec("general") {
        val allowedUploadExtensions by required<String>("allowed_upload_extensions")
        val wikiUrl by required<String>("wiki_url")
        val backupInterval by required<Long>("backup_interval")
    }

    object General {
        val allowedUploadExtensions by c(GeneralSpec.allowedUploadExtensions)
        val allowedUploadExtensionSet by lazy {
            allowedUploadExtensions.split(",").map { it.trim().toLowerCase() }.toSet()
        }
        val wikiUrl by c(GeneralSpec.wikiUrl)
        val backupInterval by c(GeneralSpec.backupInterval)
    }

    private object TwitterSpec : ConfigSpec("twitter") {
        val timeline by required<String>("timeline")
    }

    object Twitter {
        val timeline by c(TwitterSpec.timeline)
    }

    private object ResoSpec : ConfigSpec("reso") {
        val day by required<Int>()
        val time by required<Int>()
    }

    object Reso {
        val day by c(ResoSpec.day)
        val time by c(ResoSpec.time)
    }

    init {
        var config = Config {
            addSpec(ServerSpec)
            addSpec(PathSpec)
            addSpec(ScheduleSpec)
            addSpec(SecuritySpec)
            addSpec(GeneralSpec)
            addSpec(TwitterSpec)
            addSpec(ResoSpec)
        }.from.toml.resource("portal.toml")

        for (file in Files.list(Paths.get("."))) {
            if (file.fileName.toString().endsWith(".toml")) {
                try {
                    config = config.from.toml.file(file.toFile())
                } catch (_: FileNotFoundException) {
                }
            }
        }

        this.config = config.from.env()
            .from.systemProperties()
    }
}